# pylint: disable=no-self-use
from flask import request
from flask_restplus import Namespace, Resource, inputs, abort
from flask_restplus.reqparse import RequestParser
from spectacular import FailedToValidate

from airflow import settings
from shu.interactions import Interactions, InteractionError
from shu.wrappers import DagRunWrapper
from shu.adapters import create_adapters

dags = Namespace("dags")

# Explanation of priorities (#A, #B, #C)
#  #A: Needs to be in prototype
#  #B: Needs to be in when deploying to production
#  #C: Better behaviour, but at this stage non-essential

# IMPR: [#C] Ensure the DAG model also appears in the API documentation,
#            Add something like, marshall_with to DagCollection.get
# DONE: [#B] Better responses on failure
#            Check generic error handler, make handler for PersistenceHandler


class Parser:
    @staticmethod
    def get_dag_collection():
        parser = RequestParser()
        parser.add_argument("prefix", type=str, required=False)
        parser.add_argument("is_active", type=inputs.boolean)
        return parser

    @staticmethod
    def get_dag_run_collection():
        parser = RequestParser()
        parser.add_argument("prefix", type=str, required=False)
        parser.add_argument("state", type=str, required=False)
        return parser


def interactions():
    adapters = create_adapters(settings.Session())
    return Interactions(adapters)


@dags.route("")
class DagCollection(Resource):
    @dags.expect(Parser.get_dag_collection())
    def get(self):
        """List DAGs"""
        try:
            arguments = Parser.get_dag_collection().parse_args()
            return interactions().find_dags(arguments).to_dict()
        except (InteractionError, FailedToValidate) as exc:
            abort(400, error=exc.to_dict())


@dags.route("/<string:dag_id>/runs")
class DagRunCollection(Resource):
    @dags.expect(dags.schema_model("DagRun", DagRunWrapper.schema(ignored=["dag_id"])))
    def post(self, dag_id):
        try:
            return interactions().create_dag_run(dag_id, request.json).to_dict()
        except (InteractionError, FailedToValidate) as exc:
            abort(400, error=exc.to_dict())

    @dags.expect(Parser.get_dag_run_collection())
    def get(self, dag_id):
        """List DAG runs"""
        try:
            return interactions().find_dag_run(dag_id, request.args).to_dict()
        except (InteractionError, FailedToValidate) as exc:
            abort(400, error=exc.to_dict())



import datetime as dt
from abc import ABC, abstractclassmethod, abstractmethod

import spectacular as spec
from airflow.models import DagRun


class Wrapper(ABC):
    @abstractmethod
    def to_dict(self):
        pass

    @abstractclassmethod
    def from_dict(cls, data):
        pass


class Collection:
    def __init__(self, xs):
        self.xs = xs

    def to_dict(self):
        return [x.to_dict() for x in self.xs]


def from_schedule_interval(schedule_interval):
    if isinstance(schedule_interval, dt.timedelta):
        return {"type": "seconds", "value": schedule_interval.total_seconds()}
    if isinstance(schedule_interval, str):
        return {"type": "cron", "value": schedule_interval}
    raise ValueError("Expected either a string or a timedelta")


def from_state(state):
    return state


class DagWrapper(Wrapper):
    def __init__(self, dag_model):
        self.dag_model = dag_model

    def to_dict(self):
        return {
            "dag_id": self.dag_model.dag_id,
            "is_paused": self.dag_model.is_paused,
            "is_active": self.dag_model.is_active,
            "schedule_interval": from_schedule_interval(
                self.dag_model.schedule_interval
            ),
        }

    def from_dict(self, data):
        raise NotImplementedError()


class DagRunWrapper(Wrapper):
    def __init__(self, dag_run):
        self.dag_run = dag_run

    @staticmethod
    def schema(context=spec.Context.CREATE, ignored=None):
        conf_spec = spec.any_value()
        conf_spec["example"] = {}
        schema = spec.obj(conf=conf_spec, run_id=spec.string(), dag_id=spec.string())
        return spec.select(schema, context, ignored=ignored or [])

    @classmethod
    def from_dict(cls, data):
        spec.validate(cls.schema(), data)
        # pylint: disable=unexpected-keyword-arg
        dag_run = DagRun(
            dag_id=data["dag_id"],
            run_id=data["run_id"],
            conf=data["conf"],
            external_trigger=True,
        )
        return cls(dag_run)

    def to_dict(self):
        return {
            "dag_id": self.dag_run.dag_id,
            "run_id": self.dag_run.run_id,
            "conf": self.dag_run.conf,
            # pylint: disable=protected-access; HACK
            "state": from_state(self.dag_run._state),
        }



from airflow.models import DagModel, DagRun


class RepoTransaction:
    def __init__(self, session):
        self._session = session

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self._session.commit()
        else:
            self._session.rollback()
        return False  # Do not surpress the exception


class DagRunRepository:
    def __init__(self, session):
        self._session = session

    def transaction(self):
        return RepoTransaction(self._session)

    def store(self, run):
        self._session.add(run)

    def find(self, dag_id, filters):
        query = self._session.query(DagRun).filter(DagRun.dag_id == dag_id)
        if "prefix" in filters:
            pattern = "{}%".format(filters["prefix"])
            query = query.filter(DagRun.run_id.like(pattern))
            del filters["prefix"]
        return query.filter_by(**filters).all()

    def get(self, dag_id, run_id):
        query = self._session.query(DagRun).filter(
            DagRun.dag_id == dag_id, DagRun.run_id == run_id
        )
        return query.one_or_none()


class DagRepository:
    def __init__(self, session):
        self._session = session

    def transaction(self):
        return RepoTransaction(self._session)

    def find(self, filters):
        query = self._session.query(DagModel)
        if "prefix" in filters:
            pattern = "{}%".format(filters["prefix"])
            query = query.filter(DagModel.dag_id.like(pattern))
            del filters["prefix"]
        return query.filter_by(**filters).all()

    def get(self, dag_id):
        query = self._session.query(DagModel).filter(DagModel.dag_id == dag_id)
        return query.one_or_none()


def create_adapters(session):
    return {
        "dag_run_repo": DagRunRepository(session),
        "dag_repo": DagRepository(session),
    }



import logging
from airflow.models import DAG, DagRun
from shu.wrappers import DagRunWrapper, DagWrapper, Collection


log = logging.getLogger(__name__)


class InteractionError(Exception):
    pass


class NotFound(InteractionError):
    def __init__(self, cls, key):
        super().__init__("No {} with key {}".format(cls.__name__, key))
        self.cls = cls
        self.key = key

    def to_dict(self):
        return {"object": self.cls.__name__, "key": self.key, "message": str(self)}


class AlreadyExists(InteractionError):
    def __init__(self, cls, key):
        super().__init__("Key {} for {} already exists".format(key, cls.__name__))
        self.cls = cls
        self.key = key

    def to_dict(self):
        return {"object": self.cls.__name__, "key": self.key, "message": str(self)}


class Interactions:
    def __init__(self, adapters):
        self._dag_run_repo = adapters["dag_run_repo"]
        self._dag_repo = adapters["dag_repo"]

    def find_dags(self, filters):
        cleaned = {k: v for (k, v) in filters.items() if v is not None}
        dags = self._dag_repo.find(cleaned)
        return Collection([DagWrapper(dag) for dag in dags])

    def create_dag_run(self, dag_id, data):
        dag = self._dag_repo.get(dag_id)
        if not dag:
            raise NotFound(DAG, dag_id)
        data["dag_id"] = dag_id

        wrapper = DagRunWrapper.from_dict(data)
        existing = self._dag_run_repo.get(dag_id, wrapper.dag_run.run_id)
        if existing is not None:
            raise AlreadyExists(DagRun, wrapper.dag_run.run_id)
        with self._dag_run_repo.transaction():
            self._dag_run_repo.store(wrapper.dag_run)
        return wrapper

    def find_dag_run(self, dag_id, filters):
        cleaned = {k: v for (k, v) in filters.items() if v is not None}
        dag_runs = self._dag_run_repo.find(dag_id, cleaned)
        return Collection([DagRunWrapper(dag_run) for dag_run in dag_runs])



# This is the class you derive to create a plugin
import logging

from airflow.plugins_manager import AirflowPlugin
from airflow.www.app import csrf

from shu.api import create_api, create_swagger
from shu.resources import dags

log = logging.getLogger(__name__)


def create_blueprints():
    api = create_api("/internal")
    api.add_namespace(dags)
    swagger = create_swagger()
    csrf.exempt(api.blueprint)
    csrf.exempt(swagger)
    return [api.blueprint, swagger]


# Defining the plugin class
class Plugin(AirflowPlugin):
    name = "internal"
    operators = []
    sensors = []
    hooks = []
    executors = []
    macros = []
    admin_views = []
    flask_blueprints = create_blueprints()
    menu_links = []
    appbuilder_views = []
    appbuilder_menu_items = []
    global_operator_extra_links = []



import logging
import os

from flask import Blueprint, url_for
from flask_restplus import Api, Resource, apidoc, Namespace
from flask_restplus.api import SwaggerView

log = logging.getLogger(__name__)


def create_api(prefix, **extra):
    # pylint: disable=unused-variable
    """Create an API with info and health endpoints.

    :param name: The name of the application; used in the generated frontend,
        and the URL prefix.
    :param config: The application configuration.

    :returns: An instance of :class:`~apian.api.CustomAPI`.
    """
    blueprint = Blueprint(prefix, __name__)
    api = CustomAPI(blueprint, title="api", prefix=prefix, doc=prefix + "/", **extra)

    ns = Namespace("Generic")

    @api.route("/health")
    class Health(Resource):
        def get(self):
            # pylint: disable=no-self-use
            """Check the health of the service"""
            return {"status": "ok"}

    return api


def create_swagger():
    swagger = apidoc.Apidoc(
        "restplus_custom_doc",
        __name__,
        template_folder="templates",
        static_folder=os.path.dirname(apidoc.__file__) + "/static",
        static_url_path="/swagger",
    )

    @swagger.add_app_template_global
    def swagger_static(filename):
        # pylint: disable=unused-variable
        return url_for("restplus_custom_doc.static", filename=filename)

    return swagger


class CustomAPI(Api):
    """
    Hack to serve swagger.json from the /api prefix.  This is a requirement for
    serving swagger-ui from behind a path-routing load balancer without
    additional proxies.
    """

    def _register_specs(self, app_or_blueprint):
        if self._add_specs:
            endpoint = str("specs")
            self._register_view(
                app_or_blueprint,
                SwaggerView,
                self.default_namespace,
                "{}/swagger.json".format(self.prefix),
                endpoint=endpoint,
                resource_class_args=(self,),
            )
            self.endpoints.add(endpoint)

    @property
    def specs_url(self):
        """Swagger URI protocol-relative."""
        return url_for(self.endpoint("specs"), _external=True, _scheme="")
